import sys
from copy import copy
from dataclasses import dataclass
from difflib import SequenceMatcher
from itertools import tee
from pathlib import Path
from textwrap import dedent
from typing import List

import typer
from tabulate import tabulate
from typer import Argument, Option

from nef_pipelines.lib.nef_lib import (
    NEF_MOLECULAR_SYSTEM,
    SELECTORS_LOWER,
    SelectionType,
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames,
)
from nef_pipelines.lib.sequence_lib import (
    chains_from_frames,
    get_chain_starts,
    select_best_residues_by_info_content,
    sequence_to_chains,
    sequence_to_residue_name_lookup,
    sequences_from_frames,
)
from nef_pipelines.lib.util import (
    chunks,
    end_with_ordinal,
    exit_error,
    is_int,
    parse_comma_separated_options,
    strings_to_tabulated_terminal_sensitive,
)
from nef_pipelines.tools.chains import chains_app
from nef_pipelines.tools.chains.renumber import pipe as renumber

REFERENCE_FRAMES_HELP = """
frames to find the starting residue of a chain for renumbering when using starts  you can have multiple reference
frames and the lowest residue number found amongst them is used. The option can be repeated  or you can use comma
separated lists of frame names [default: the molecular system frame if present otherwise all frames].
"""

app = typer.Typer()

CHAIN_CODE = "chain_code"

SEQUENCE_CODE = "sequence_code"


@dataclass
class Sequence:
    chain_code: str
    start: int
    sequence: List[str]


@chains_app.command()
def align(
    reference_chains_and_frames: List[str] = Argument(
        None,
        metavar="<REFERENCE-CHAIN-CODE> " "<REFERENCE-FRAME-SELECTOR>",
        help="chain-codes and frame names to align the chain to. To "
        "disambiguate chains from frames add an # to the front of the chain"
        "[default: align all chains in all frames to molecular_system "
        "frame chains",
        show_default=False,
    ),
    reference_selector_type: SelectionType = Option(
        SelectionType.ANY,
        "--reference-selector-type",
        help=f"choose how to select the reference frames to align to, "
        f"can be one of: {SelectionType.__members__.keys()}. ",
    ),
    input: Path = typer.Option(
        Path("-"),
        "-i",
        "--in",
        metavar="NEF-FILE",
        help="where to read NEF data from either a file or stdin '-'",
    ),
    frame_selectors: List[str] = Option(
        None,
        "-f",
        "--frames",
        metavar="<FRAME-SELECTOR>",
        help="Limit changes to a particular frame by name or category [the selector], note: wildcards are "
        "allowed. Frames are selected by name and then category if the name doesn't match"
        "note: the option -t /--selector-type allows you to choose which selection type to use. "
        "[default: all frames]",
        show_default=False,
    ),
    frame_selector_type: SelectionType = typer.Option(
        SelectionType.ANY,
        "-s",
        "--selector-type",
        help=f"control how to select frames to renumber, can be one of: {SELECTORS_LOWER}. "
        "Any will match on names first and then if there is no match attempt to match on category",
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        help="display verbose alignment information [default: false]",
    ),
):
    """- align chains in the frames in a nef file so they match those in a reference frame [alpha]"""

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    reference_chains_and_frames = parse_comma_separated_options(
        reference_chains_and_frames
    )

    reference_chains, reference_frames = (
        _parse_target_chains_and_reference_frames_or_exit_error(
            reference_chains_and_frames, entry, reference_selector_type
        )
    )

    reference_sequences = sequences_from_frames(reference_frames)

    reference_sequences_by_chains = _build_per_chain_alignment_sequences(
        reference_sequences, reference_chains
    )

    if not frame_selectors:
        frame_selectors = ["*"]

    target_frames = select_frames(entry, frame_selectors, frame_selector_type)

    target_frames = [frame for frame in target_frames if frame not in reference_frames]

    target_sequence_by_frames_and_chains = {}
    for target_frame in target_frames:
        target_sequence = sequences_from_frames(target_frame)
        target_sequence = [
            residue
            for residue in target_sequence
            if residue.chain_code in reference_chains and residue.residue_name
        ]

        target_sequences = _build_per_chain_alignment_sequences(
            target_sequence, reference_chains
        )

        if not target_sequences:
            continue

        target_sequence_by_frames_and_chains[target_frame.name] = target_sequences

    for (
        target_frame_name,
        target_sequences_by_chain,
    ) in target_sequence_by_frames_and_chains.items():
        for target_chain, target_sequence in target_sequences_by_chain.items():

            reference_sequence = reference_sequences_by_chains[target_chain]

            if (not reference_sequence.sequence) or (not target_sequence.sequence):
                continue

            matcher = SequenceMatcher(
                a=reference_sequence.sequence,
                b=target_sequence.sequence,
                autojunk=False,
            )

            if verbose:
                entry_id = entry.entry_id
                ref_chain_code = reference_sequence.chain_code
                ratio = f"{matcher.ratio():7.3f}"
                print(
                    f"[{entry_id}] align chain {target_chain} {target_frame_name} -> {ref_chain_code} ratio: {ratio}",
                    file=sys.stderr,
                )

            offset = _get_offset_or_none(matcher)

            _warn_if_no_match(offset, reference_frames, target_frame)

            if offset is None:
                continue

            entry = renumber(
                entry,
                target_frame_name,
                SelectionType.NAME,
                {
                    target_chain: reference_sequence.start
                    - 1
                    - target_sequence.start
                    + offset
                    + 1
                },
            )

    print(entry)


def _get_offset_or_none(matcher):
    offset = None
    op_codes = matcher.get_opcodes()

    equal_elem_offsets_by_length = {}
    for elem in op_codes:
        if elem[0] in ("replace", "delete", "insert"):
            continue
        elif elem[0] == "equal":
            offset = elem[1] - elem[3]
            equal_elem_offsets_by_length[elem[2] - elem[1]] = offset
            continue

    if equal_elem_offsets_by_length:
        offset = equal_elem_offsets_by_length[max(equal_elem_offsets_by_length)]
    return offset


#
# def pipe(
#     entry: Entry,
#     chains: List[str],
#     reference_frames: List[Saveframe],
#     target_frames: List[Saveframe],
# ) -> Entry:
#
#     frames = select_frames(entry, frame_selectors, frame_selector_type)
#     #
#     # _exit_if_selected_chain_not_in_frames(
#     #     "chain to renumber", frames, input, offsets.keys()
#     # )
#     #
#     # offset_chains_in_frames(frames, offsets)
#
#     return entry


def align_seqs(seq, newseq, wildcard="***"):
    for i in range(len(seq)):
        if seq[i] != newseq[i]:
            if seq[i] == wildcard:
                newseq.insert(i, wildcard)

            elif newseq[i] == wildcard:
                newseq.insert(i, seq[i])
            else:
                newseq.insert(i, seq[i])

    else:
        newseq = newseq[0 : len(seq)]

    return newseq


def _warn_if_no_match(offset, reference_frames, target_frame):
    if offset is None:
        reference_frame_names = ", ".join(
            [reference_frame.name for reference_frame in reference_frames]
        )
        msg = f"""
                WARNING: couldn't align the frame {target_frame.name} to the sequences in {reference_frame_names}
                         this frame was ignored
                """
        msg = dedent(msg)
        print(msg, file=sys.stderr)


def _build_per_chain_alignment_sequences(target_sequence, reference_chains):
    sequence_codes_and_names = sequence_to_residue_name_lookup(target_sequence)
    sequence_codes_and_names = {
        (chain, int(sequence_code)): residue_name
        for (chain, sequence_code), residue_name in sequence_codes_and_names.items()
        if is_int(sequence_code)
    }
    sequence_chains = sequence_to_chains(target_sequence)
    chain_sequence_codes_and_names = {}
    for chain_code in sequence_chains:
        chain_sequence_codes_and_names[chain_code] = {
            sequence_code: residue_name
            for (
                target_chain_code,
                sequence_code,
            ), residue_name in sequence_codes_and_names.items()
            if target_chain_code == chain_code
        }

    reference_sequences_by_chain = {}
    for chain in reference_chains:
        if chain not in chain_sequence_codes_and_names:
            continue
        sequence_codes_to_residue_names = {
            sequence_code: residue_name
            for sequence_code, residue_name in chain_sequence_codes_and_names[
                chain
            ].items()
        }
        sequence_codes = set(sequence_codes_to_residue_names.keys())

        min_sequence_code = min(sequence_codes)
        max_sequence_code = max(sequence_codes) + 1

        sequence_letters = []
        sequence_codes = set(sequence_codes)
        for i in range(min_sequence_code, max_sequence_code):
            residue = (
                sequence_codes_to_residue_names[i] if i in sequence_codes else "***"
            )
            sequence_letters.append(residue)

        sequence = Sequence(chain, min_sequence_code, sequence_letters)
        reference_sequences_by_chain[chain] = sequence

    return reference_sequences_by_chain


def _exit_if_selected_chain_not_in_frames(chain_type, frames, input, chains):
    frame_chains = set()
    for frame in frames:
        frame_chains.update(chains_from_frames(frame))
    frame_chain_names = strings_to_tabulated_terminal_sensitive(frame_chains)

    for chain in chains:

        selected_frames = strings_to_tabulated_terminal_sensitive(
            [frame.name for frame in frames]
        )
        if chain not in frame_chains:
            msg = """
                using the input {input}
                the {chain_type} '{chain}' is not in the frames selected for renumbering

                the selected frames are:

                {selected_frames}

                which have the chains

                {frame_chain_names}
                """
            msg = msg.format(
                chain_type=chain_type,
                input=input,
                chain=chain,
                selected_frames=selected_frames,
                frame_chain_names=frame_chain_names,
            )
            exit_error(msg)


def _exit_error_if_no_chain_offsets(chain_offsets, reference_frames, input):
    if not chain_offsets:
        sequences = [sequences_from_frames(frame) for frame in reference_frames]

        all_residues = set()
        for sequence in sequences:
            all_residues.update(sequence)

        all_residues = select_best_residues_by_info_content(all_residues)

        chain_starts = [
            [chain, start] for chain, start in get_chain_starts(all_residues).items()
        ]

        reference_frame_names = strings_to_tabulated_terminal_sensitive(
            [frame.name for frame in reference_frames]
        )

        if chain_starts:
            headings = "chain start".split()
            table = tabulate(chain_starts, headings)
            table_msg = (
                "here are the starts for the chains in the selected reference frames"
            )
        else:
            table = ""
            table_msg = (
                "also no chain starts were found in the selected reference frames"
            )

        msg = """
                in the file {input}
                you didn't provide any chains and offsets/starts
                {table_msg}

                {table}

                the reference frames for sequences and chain starts were

                {reference_frame_names}
            """
        msg = dedent(msg)
        msg = msg.format(
            input=input,
            table=table,
            table_msg=table_msg,
            reference_frame_names=reference_frame_names,
        )

        exit_error(msg)


def _exit_error_if_no_reference_frames(
    entry, reference_frame_selectors, reference_frames
):
    if not reference_frames:
        msg = """
            no reference frames found using the selection:

            {reference_frame_selectors}

            possible frame names are

            {frame_names}
            """
        msg = dedent(msg)
        reference_frame_selectors = ", ".join(reference_frame_selectors)
        reference_frame_selectors = strings_to_tabulated_terminal_sensitive(
            reference_frame_selectors
        )

        frame_names = strings_to_tabulated_terminal_sensitive(
            [frame.name for frame in entry.frame_list]
        )
        msg = msg.format(
            reference_frame_selectors=reference_frame_selectors, frame_names=frame_names
        )

        exit_error(msg)


def _select_reference_frames(entry, reference_frame_selectors):

    if not reference_frame_selectors:
        reference_frames = select_frames(
            entry, NEF_MOLECULAR_SYSTEM, SelectionType.CATEGORY
        )

        if not reference_frames:
            reference_frames = copy(entry.frame_list)
    else:
        reference_frames = []
        for frame_selector in reference_frame_selectors:
            reference_frames += select_frames(entry, frame_selector, SelectionType.ANY)

    return reference_frames


def _chain_starts_to_offsets(current_chains_starts, new_chain_starts):
    result = {}

    for chain, new_start in new_chain_starts.items():
        old_start = current_chains_starts[chain]

        offset = new_start - old_start
        result[chain] = offset

    return result


def exit_if_molecular_system_isnt_a_singleton(molecular_system_frames):
    number_of_molecular_system_frames = len(molecular_system_frames)
    if number_of_molecular_system_frames == 0:
        exit_error("there must be a molecular system defined to use the starts option")
    elif number_of_molecular_system_frames > 1:
        # could go ahead and combine but warn of inconsistent results
        exit_error(
            "there can only be one  molecular system frame defined in a NEF file"
        )


def offset_chains_in_frames(frames, chain_offsets):
    for frame in frames:
        for chain, offset in chain_offsets.items():
            _offset_residue_numbers(frame, chain, offset)


def _exit_multiple_offsets_for_chain(chain, first_offset, second_offset):
    msg = """
        the chain {chain} has more than one offset/start to use for renumbering

        the first offset/start is {first_offset}
        the second offset/start is {second_offset}
        """
    msg = msg.format(
        input=input, chain=chain, first_offset=first_offset, second_offset=second_offset
    )

    exit_error(msg)


def _get_chain_offset_pairs_or_exit_error(chain_offsets):

    # tee here as we use the chain_offsets generator twice
    chain_offsets_check, chain_offsets = tee(chunks(chain_offsets, 2))

    for chain_offset in chain_offsets_check:
        if len(chain_offset) % 2:
            chain_offset_pairs = []

            for chain_offset in chain_offsets:
                chain_offset_pairs.append(":".join(chain_offset))

            msg = f"""\
                there must be an offset/start for each chain
                i got the following pairs of chain:offset/start: {', '.join(chain_offset_pairs)}
                the last one is missing an offset.
            """
            exit_error(msg)

    result = {}
    seen_chains = set()
    for i, (chain, offset) in enumerate(chain_offsets, start=1):

        _exit_if_offset_isnt_int(chain, offset, i)

        offset = int(offset)

        if chain in seen_chains:
            _exit_multiple_offsets_for_chain(
                chain=chain, first_offset=result[chain], second_offset=offset
            )

        seen_chains.add(chain)
        result[chain] = offset

    return result


def _exit_if_offset_isnt_int(chain, offset, i):
    if not is_int(offset):
        msg = f"""\
                the offset/start {offset} in the {end_with_ordinal(i)} chain offset/start pair: {chain} {offset}
                can't be converted to an int
            """
        exit_error(msg)


def _exit_if_chains_and_offsets_dont_match(chains, offsets):
    num_chains = len(chains)
    num_offsets = len(offsets)
    if num_chains != num_offsets:
        msg = f"""\
            chains and offsets must have the same length i got {num_chains} chains and {num_offsets}
            chains: {','.join(chains)}
            offsets: {','.join([str(offset) for offset in offsets])}
        """
        exit_error(msg)


def _offset_residue_numbers(frame, chain, offset):

    for loop_data in frame.loop_dict.values():
        for tag in loop_data.tags:
            if _tag_is_based_on(tag, SEQUENCE_CODE):
                sequence_index = loop_data.tag_index(tag)

                chain_code_tag = _tag_based_on(tag, SEQUENCE_CODE, CHAIN_CODE)
                chain_code_index = loop_data.tag_index(chain_code_tag)
                for line in loop_data:
                    if chain_code_index is not None and line[chain_code_index] == chain:
                        sequence = line[sequence_index]
                        if is_int(sequence):
                            sequence = int(sequence)
                            sequence = sequence + offset
                            line[sequence_index] = str(sequence)


def _tag_based_on(tag, base, substitute):
    column_code = tag[len(base) :]
    new_tag = f"{substitute}{column_code}"
    return new_tag


def _tag_is_based_on(tag, base):
    return tag.startswith(base)


# noinspection PyUnusedLocal
def _parse_target_chains_and_reference_frames_or_exit_error(
    reference_chains_and_frames, entry, reference_selector_type
):

    sequences = sequences_from_frames(entry.frame_list)
    chains = set(sequence_to_chains(sequences))

    chains = set()
    frames = {}

    if not reference_chains_and_frames:
        selected_frames = select_frames(
            entry, NEF_MOLECULAR_SYSTEM, SelectionType.CATEGORY
        )
        frames = {frame.name: frame for frame in selected_frames}
        chains = chains_from_frames(frames.values())
    else:
        for elem in reference_chains_and_frames:

            if elem.startswith("#"):
                chain = elem.lstrip("#")
                if chain not in chains:
                    msg = f"""
                        when selecting chains to align to, the chain {chain} is not in the chains in the read frames
                        {', '.join(sorted(chains))}
                        """
                    exit_error(msg)
                chains.add(elem)
                continue
            elif elem in chains:
                chains.add(elem)
                continue

            for frame in select_frames(entry, elem, reference_selector_type):
                frames[frame.name] = frame

    chains = sorted(chains)

    return chains, frames.values()  # chains = chains_from_entry(entry)



from fnmatch import fnmatch
from pathlib import Path
from typing import List

import typer

from nef_pipelines.lib.nef_lib import read_entry_from_file_or_stdin_or_exit_error
from nef_pipelines.tools.frames import frames_app

UNDERSCORE = "_"

parser = None


# noinspection PyUnusedLocal
@frames_app.command()
def delete(
    input_path: Path = typer.Option(
        None,
        metavar="|PIPE|",
        help="file to read NEF data from default is stdin '-'",
    ),
    use_categories: bool = typer.Option(
        False,
        "-c",
        "--category",
        help="if selected use the category of the frame to select it for deletion rather than it name",
    ),
    exact: bool = typer.Option(
        False, "-e", "--exact", help="don't treat name as a wild card"
    ),
    selectors: List[str] = typer.Argument(
        ...,
        help="a list of frames to delete by type or name,  names can be wildcards, names have lead _'s removed and "
        "surrounding back quotes `  removed",
    ),
):
    """- delete frames in the current input by type or name"""

    entry = read_entry_from_file_or_stdin_or_exit_error(input_path)

    to_delete = []
    for name in selectors:
        for frame in entry:
            frame_full_name = frame.name
            frame_category = frame.category
            frame_name = frame_full_name[len(frame_category) :].lstrip("_").strip("`")

            for selector in selectors:
                if not exact:
                    selector = f"*{selector}*"

                if use_categories:
                    if fnmatch(frame_category, selector):
                        to_delete.append(frame)
                else:
                    if fnmatch(frame_name, selector):
                        to_delete.append(frame)

    entry.remove_saveframe(to_delete)

    print(entry)



import typer

import nef_pipelines
from nef_pipelines import nef_app

app = typer.Typer()
import_app = typer.Typer()
export_app = typer.Typer()

if nef_app.app:

    nef_app.app.add_typer(app, name="fasta", help="- read and write fasta sequences")

    app.add_typer(import_app, name="import", help="- import fasta sequences")
    app.add_typer(export_app, name="export", help="- export fasta sequences")

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.transcoders.fasta.exporters.sequence  # noqa: F401
    import nef_pipelines.transcoders.fasta.importers.sequence  # noqa: F401



# TODO: support NMRPipe his and cis variants
# TODO: what do unassigned nmrpipe shifts look like?
# TODO: support general nmrpipe shift input
# TODO: support the HA2|HA3 syntax
# TODO: support unassigned
# TODO: support multiple chains & files
from contextlib import redirect_stdout
from io import StringIO
from pathlib import Path
from typing import List

import tabulate
import typer
from click import get_current_context
from pynmrstar import Entry

from nef_pipelines.lib.nef_frames_lib import SHIFT_LIST_FRAME_CATEGORY
from nef_pipelines.lib.nef_lib import (
    get_frame_id,
    get_frame_ids,
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames_by_name,
)
from nef_pipelines.lib.sequence_lib import (
    TRANSLATIONS_3_1,
    MoleculeTypes,
    get_chain_starts,
    make_chunked_sequence_1let,
    sequence_from_entry,
    sequence_residues_to_sequence_3let,
    sequence_to_chains,
    translate_3_to_1,
)
from nef_pipelines.lib.shift_lib import nef_frames_to_shifts
from nef_pipelines.lib.util import STDIN, exit_error, is_float, is_int
from nef_pipelines.transcoders.talos import export_app

app = typer.Typer()


# noinspection PyUnusedLocal
@export_app.command()
def shifts(
    user_chain: str = typer.Option(
        None,
        "-c",
        "--chain",
        help="chain to export",
        metavar="<CHAIN-CODE>",
    ),
    output_to_stdout: bool = typer.Option(
        False, "-o", "--out", help="write the files to stdout for debugging"
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--input",
        help="input to read NEF data from [- is stdin]",
    ),
    frame_selector: str = typer.Argument(
        None,
        help="selector for the chemical shift list frame to use",
    ),
):
    """- convert nef shifts (and sequence) to talos input [alpha]"""

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    user_chain = _select_chain_or_exit(entry, user_chain)

    output = pipe(entry, user_chain, frame_selector)

    print("\n".join(output))


def pipe(entry: Entry, target_chain_code: str, frame_selector: str) -> List[str]:

    target_shift_frame = _select_shift_frame_or_exit(entry, frame_selector)

    sequence = sequence_from_entry(entry)

    target_sequence = [
        residue for residue in sequence if residue.chain_code == target_chain_code
    ]

    start_residue = get_chain_starts(target_sequence)[target_chain_code]

    sequence_1let = translate_3_to_1(
        sequence_residues_to_sequence_3let(target_sequence)
    )

    for i, residue in enumerate(target_sequence):
        if _is_talos_variant(residue.variants, residue.residue_name):
            sequence_1let[i] = sequence_1let[i].lower()

    chunked_sequence = make_chunked_sequence_1let(sequence_1let, line_length=50)

    shifts = nef_frames_to_shifts(
        [target_shift_frame],
    )

    shifts = [
        shift for shift in shifts if shift.atom.residue.chain_code == target_chain_code
    ]

    TALOS_ATOMS = set("H N CA CB C HA HA2 HA3".split())
    shifts = [shift for shift in shifts if shift.atom.atom_name in TALOS_ATOMS]

    shifts = [shift for shift in shifts if is_int(shift.atom.residue.sequence_code)]

    shifts = [shift for shift in shifts if is_float(shift.value)]

    shifts = [
        shift
        for shift in shifts
        if shift.atom.residue.residue_name in TRANSLATIONS_3_1[MoleculeTypes.PROTEIN]
    ]

    with redirect_stdout(StringIO()) as capture:
        print(f"REMARK Chemical shift table for {entry.entry_id}")
        print()
        print(f"DATA CHAIN {target_chain_code}")
        print(f"DATA FIRST_RESID {start_residue}")
        print()
        for sub_sequence in chunked_sequence:
            print(f"DATA SEQUENCE {sub_sequence}")
        print()

        headers = [
            "VARS   RESID\nFORMAT %4d",
            "RESNAME\n%1s",
            "ATOMNAME\n%4s",
            "SHIFT\n%8.3f",
        ]

        table = []

        for j, shift in enumerate(shifts):
            sequence_offset = shift.atom.residue.sequence_code - start_residue
            residue_name_3let = shift.atom.residue.residue_name
            residue_name_1let = translate_3_to_1(
                [
                    residue_name_3let,
                ]
            )[0]

            if _is_talos_variant(sequence[sequence_offset].variants, residue_name_3let):
                residue_name_1let = residue_name_1let.lower()

            table.append(
                [
                    int(shift.atom.residue.sequence_code),
                    residue_name_1let,
                    shift.atom.atom_name,
                    float(shift.value),
                ]
            )

        print(
            tabulate.tabulate(
                table,
                tablefmt="plain",
                headers=headers,
                colalign=["left", "left", "left", "decimal"],
            )
        )

    return str(capture.getvalue()).split("\n")


def _is_talos_variant(variants, residue_name):
    return (
        "+HE2" in variants
        and residue_name == "HIS"
        or "-HG" in variants
        and residue_name == "CYS"
    )


def _select_shift_frame_or_exit(entry, frame_selector):

    shift_frames = entry.get_saveframes_by_category(SHIFT_LIST_FRAME_CATEGORY)

    if len(shift_frames) == 0:
        print(get_current_context().get_help())
        exit_error("no shift frames in input stream")

    if len(shift_frames) == 1:
        target_shift_frame = shift_frames[0]
        if frame_selector is not None:
            target_id = get_frame_id(target_shift_frame)
            if target_id != frame_selector:
                msg = f"""
                    the selected frame [{frame_selector}] is different from the frame found [{target_id}]
                    as only one shift frame is present you don't need a frame selector [--frame]!
                    """
                exit_error(msg)

    if len(shift_frames) > 1:
        if frame_selector is None:
            frame_ids = ",".join(get_frame_ids(shift_frames))
            msg = f"""
                    multiple shift frames in the stream but no frame selection option [use --frame to select one]
                    the frame ids are {frame_ids}
                """
            exit_error(msg)
        else:
            selected_frames = select_frames_by_name(shift_frames, frame_selector)
            if len(selected_frames) == 0:
                frame_ids = ", ".join(get_frame_ids(shift_frames))
                msg = f"""
                    the selector {frame_selector} didn't selected a frame FROM the available frames
                    the ids of the available shift frames are {frame_ids}
                    """
                exit_error(msg)

    return target_shift_frame


def _select_chain_or_exit(entry, user_chain):

    sequence = sequence_from_entry(entry)

    sequence_chains = sequence_to_chains(sequence)

    if len(sequence_chains) == 0:
        msg = """
            there are no chains in the molecular sequence!
        """
        exit_error(msg)

    if len(sequence_chains) == 1:
        target_sequence_chain = sequence_chains[0]
        if user_chain is not None and target_sequence_chain != user_chain:
            msg = f"""
                    the selected chain_code [{user_chain}] is not in the available chain_codes [{target_sequence_chain}]
                    as only one chain is present you don't need a chain selector [--chain]!
                """
            exit_error(msg)

    if len(sequence_chains) > 1:
        if user_chain is not None and user_chain not in sequence_chains:
            chain_codes = ", ".join(sequence_chains)
            msg = f"""
                    the selected chain_code {user_chain} is not one of the available chain_codes {chain_codes}
                """
            exit_error(msg)
        if user_chain is None:
            msg = f"""
                    multiple chains found [{chain_codes}] but no chain selector [--chain] given
                """
            exit_error(msg)
        target_sequence_chain = user_chain

    return target_sequence_chain



from argparse import Namespace
from itertools import zip_longest
from pathlib import Path
from typing import List

import typer

from nef_pipelines.lib.nef_lib import file_name_path_to_frame_name
from nef_pipelines.lib.sequence_lib import (
    ANY_CHAIN,
    get_sequence_or_exit,
    sequence_to_residue_name_lookup,
)
from nef_pipelines.lib.util import (
    parse_comma_separated_options,
    process_stream_and_add_frames,
)
from nef_pipelines.transcoders.xplor import import_app
from nef_pipelines.transcoders.xplor.xplor_lib import (
    _exit_if_chains_and_filenames_dont_match,
    dihedral_restraints_to_nef,
    read_dihedral_restraints_or_exit_error,
)

CHAINS_HELP = """\
                 A list of chains to be used instead or as well as segids. Multiple calls can be used to add multiple
                 chains. Also comma separated list of chains [without spaces] can be added [e.g. --chains A,B,C].
                 Chains are added to save frames in the order the files with restraints are added. '_' indicates to use
                 the segid in the file if chains are not it is assumed they willl be provided by segids'
              """
USE_CHAINS_HELP = (
    "use chains from the command line replacing the segids found in the file"
)


@import_app.command(no_args_is_help=True)
def dihedrals(
    input: Path = typer.Option(
        Path("-"),
        "--input",
        "-i",
        metavar="|INPUT|",
        help="input to read NEF data from [stdin = -]",
    ),
    chains: List[str] = typer.Option(
        ["A"], "--chains", "-c", metavar="<CHAIN> [,<CHAIN>,...", help=CHAINS_HELP
    ),
    use_chains: bool = typer.Option(False, "--use-chains", help=USE_CHAINS_HELP),
    file_names: List[Path] = typer.Argument(
        ..., help="input dihedral restraint tables", metavar="<DIHEDRALS-TBL>"
    ),
):
    """- read xplor dihedral restraints"""

    chains = parse_comma_separated_options(chains)

    sequence = get_sequence_or_exit(input)

    residue_name_lookup = sequence_to_residue_name_lookup(sequence)

    dihedral_restraints = []

    _exit_if_chains_and_filenames_dont_match(chains, file_names)

    frame_names = []
    for file_name, chain in zip_longest(file_names, chains, fillvalue=ANY_CHAIN):
        frame_names.append(file_name_path_to_frame_name(file_name))
        dihedral_restraints.append(
            read_dihedral_restraints_or_exit_error(
                file_name, residue_name_lookup, chain, use_chains
            )
        )

    for i, (restraint_list, frame_name) in enumerate(
        zip(dihedral_restraints, frame_names)
    ):
        nef_restraints = dihedral_restraints_to_nef(restraint_list, frame_name)

    entry = process_stream_and_add_frames(
        [nef_restraints], Namespace(pipe=None, entry_name="xplor_dihedral_restraints")
    )

    print(entry)



import itertools
import re
import string
from collections import Counter
from pathlib import Path
from textwrap import dedent
from typing import List

import typer
from fyeah import f
from pynmrstar import Entry

from nef_pipelines.lib.nef_lib import (
    add_frames_to_entry,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.util import (
    NEWLINE,
    STDIN,
    exit_error,
    parse_comma_separated_options,
)
from nef_pipelines.transcoders.nmrpipe import import_app

from ...nmrview.importers.peaks import _create_spectrum_frame
from ..nmrpipe_lib import (
    NMRPIPE_PEAK_EXPECTED_FIELDS,
    check_is_peak_file,
    get_gdb_columns,
    read_db_file_records,
    read_peak_file,
)

app = typer.Typer()

# TODO should check peaks against sequence


# noinspection PyUnusedLocal
@import_app.command(no_args_is_help=True)
def peaks(
    chain_codes: str = typer.Option(
        "A", "--chain", help="chain code", metavar="<chain-code>"
    ),
    filter_noise: bool = typer.Option(True, help="remove peaks labelled as noise"),
    in_file: Path = typer.Option(
        STDIN, "-i", "--in", help="file to read nef data from", metavar="<NEF-FILE>"
    ),
    entry_name: str = typer.Option(
        "nmrpipe", "-e", "--entry", help="entry name", metavar="<entry-name>"
    ),
    file_names: List[Path] = typer.Argument(
        ..., help="input peak files", metavar="<peak-file.xpk>"
    ),
):
    """convert nmrpipe peak file <NMRPIPE>.tab files to NEF"""

    if not chain_codes:
        chain_codes = ["A"] * len(file_names)

    if len(chain_codes) == 1:
        chain_codes = chain_codes * len(file_names)

    chain_codes = parse_comma_separated_options(chain_codes)

    _exit_if_num_chains_and_files_dont_match(chain_codes, file_names)

    entry = read_or_create_entry_exit_error_on_bad_file(in_file, entry_name=entry_name)

    entry = pipe(entry, file_names, chain_codes, filter_noise)

    print(entry)


def pipe(
    entry: Entry, file_names: List[Path], chain_codes: List[str], filter_noise: bool
) -> Entry:

    peak_lists = _read_nmrpipe_peaks(file_names, chain_codes, filter_noise=filter_noise)

    frame_name_template = "{file_name}"

    frame_names = _create_entry_names_from_template_if_required(
        frame_name_template, [], file_names
    )

    frames = []
    for i, (peak_list, frame_name, chain_code) in enumerate(
        zip(peak_lists, frame_names, chain_codes), start=1
    ):

        frames.append(_create_spectrum_frame(frame_name, peak_list, chain_code))

    # TODO when adding frames they need to be disambiguated if there are name clashes
    return add_frames_to_entry(entry, frames)


# TODO move much of this to utilities
def _exit_if_num_chains_and_files_dont_match(chain_codes, file_names):
    if len(chain_codes) != len(file_names):
        msg = f"""
            the number of chain codes {len(chain_codes)} does not match number of files {len(file_names)}
            the chain codes are {', '.join(chain_codes)}
            the files are
            {NEWLINE.join(file_names)}
        """
        exit_error(msg)


def _disambiguate_names(names):
    name_count = Counter()
    result = []
    for name in names:
        if name in name_count:
            result.append(f"{name}_{name_count[name]}")
        else:
            result.append(name)
        name_count[name] += 1
    return result


def _remove_repeated_underscores(s):
    return re.sub("_+", "_", s)


def _create_entry_names_from_template_if_required(
    entry_name_template, entry_names, file_names
):
    new_entry_names = []
    for file_name, entry_name in itertools.zip_longest(file_names, entry_names):
        if entry_name:
            new_entry_names.append(entry_name)
        else:
            file_name = str(Path(file_name).stem)
            numbers_and_letters = string.ascii_lowercase + string.digits
            file_name = "".join(  # noqa: F841
                [
                    letter if letter in numbers_and_letters else "_"
                    for letter in file_name
                ]
            )

            new_name = f(entry_name_template)

            new_name = _remove_repeated_underscores(new_name)

            new_entry_names.append(new_name)

    return _disambiguate_names(new_entry_names)


def _read_nmrpipe_peaks(file_names, chain_codes, filter_noise):
    results = []
    for file_name, chain_code in zip(file_names, chain_codes):
        with open(file_name) as file_h:
            gdb_file = read_db_file_records(file_h, file_name=file_name)

        _check_is_peak_file_or_exit(gdb_file)

        results.append(read_peak_file(gdb_file, chain_code, filter_noise=filter_noise))

    return results


def _check_is_peak_file_or_exit(gdb_file):
    columns = set(get_gdb_columns(gdb_file))

    if not check_is_peak_file(gdb_file):
        msg = f"""\
                this gdb file doesn't appear to contain all the columns expected for a peak file
                expected: {','.join(NMRPIPE_PEAK_EXPECTED_FIELDS)}
                got {','.join(columns & NMRPIPE_PEAK_EXPECTED_FIELDS)}
                file: {gdb_file.file_name}
                """
        msg = dedent(msg)
        exit_error(msg)



# original implementation by esther
import sys
from dataclasses import dataclass
from pathlib import Path
from string import digits
from textwrap import dedent
from typing import List, Union

import typer
from pynmrstar import Entry
from tabulate import tabulate

from nef_pipelines.lib.nef_lib import (
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames_by_name,
)
from nef_pipelines.lib.sequence_lib import (
    exit_if_chain_not_in_entrys_sequence,
    get_residue_name_from_lookup,
    sequence_from_entry_or_exit,
    sequence_to_residue_name_lookup,
)
from nef_pipelines.lib.shift_lib import (
    frames_to_assigned_and_unassigned_shift_lists,
    shifts_to_chains,
)
from nef_pipelines.lib.util import (
    STDIN,
    STDOUT,
    exit_error,
    exit_if_file_has_bytes_and_no_force,
    is_float,
    is_int,
)
from nef_pipelines.transcoders.mars import export_app
from nef_pipelines.transcoders.mars.util import (
    _exit_if_no_frames_selected,
    _select_target_chain_from_sequence_if_not_defined,
)

STDOUT_PATH = "-"

app = typer.Typer()

# TODO: name translations
# TODO: correct weights
# TODO: move utilities to lib

REMOVE_DIGITS = str.maketrans("", "", digits)

HEADINGS = (
    ("H", ("H", 0)),
    ("N", ("N", 0)),
    ("Ca", ("CA", 0)),
    ("Ca-1", ("CA", 1)),
    ("Cb", ("CB", 0)),
    ("Cb-1", ("CB", 1)),
    ("CO", ("C", 0)),
    ("CO-1", ("C", 1)),
    ("HA", ("HA", 0)),
    ("HA-1", ("HA", 1)),
)

OUPUT_TABLE_HEADERS = (
    "",
    "H",
    "N",
    "CA",
    "CA-1",
    "CB",
    "CB-1",
    "CO",
    "CO-1",
    "HA",
    "HA-1",
)


# TODO: replace with equivalent in structures
@dataclass(frozen=True, order=True)
class MarsAtom:
    atom_name: str
    sequence_code: Union[int, str]
    residue_name: str
    negative_offset: int = 0


# noinspection PyUnusedLocal
@export_app.command()
def shifts(
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--input",
        metavar="|PIPE|",
        help="input to read NEF data from [- is stdin]",
    ),
    target_chain: str = typer.Option(
        None,
        "-c",
        "--chain",
        help="chain to export [# and @ and the 1st chain code in the project if the is only one]",
        metavar="<CHAIN-CODE>",
    ),
    output_file: str = typer.Option(
        None,
        "-o",
        "--out",
        help="file name to output to [default <entry_id>_shifts.tab] for stdout use -",
        metavar="<MARS_SHIFT_FILE>",
    ),
    force: bool = typer.Option(
        False,
        "-f",
        "--force",
        help="force overwrite of output file if it exists and isn't empty",
    ),
    shift_frame_selectors: List[str] = typer.Argument(
        None,
        help="selector for the shift restraint frames to use, can be called multiple times and include wild cards",
    ),
):
    """- write a mars chemical shift file"""

    if not shift_frame_selectors:
        shift_frame_selectors = ["*"]

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    output_file = f"{entry.entry_id}_shifts.tab" if output_file is None else output_file

    entry = pipe(entry, shift_frame_selectors, target_chain, Path(output_file), force)

    if entry:
        print(entry)


def _assigned_shifts_filter_non_numeric_sequence_codes(assigned_shifts):
    new_assigned_shifts = set()

    for assigned_shift in assigned_shifts:
        if not is_int(assigned_shift.atom.residue.sequence_code):
            continue
        else:
            new_assigned_shifts.add(assigned_shift)

    return new_assigned_shifts


def pipe(
    entry: Entry,
    shift_frame_selectors: List[str],
    target_chain: str,
    output_file: Path,
    force: bool,
) -> Entry:

    frames = entry.get_saveframes_by_category("nef_chemical_shift_list")
    frames = select_frames_by_name(frames, shift_frame_selectors)

    _exit_if_no_frames_selected(frames)

    assigned_shifts, unassigned_shifts = frames_to_assigned_and_unassigned_shift_lists(
        frames
    )

    # TODO: filter shifts which have odd assignments
    # TODO: issue a warning if there are shifts which are not defined
    assigned_shifts = _assigned_shifts_filter_non_numeric_sequence_codes(
        assigned_shifts
    )

    shift_chains = shifts_to_chains([*assigned_shifts, *unassigned_shifts])

    target_chain = _select_target_chain_from_sequence_if_not_defined(
        target_chain, shift_chains
    )

    exit_if_chain_not_in_entrys_sequence(target_chain, entry)

    sequence = sequence_from_entry_or_exit(entry)
    residue_name_lookup = sequence_to_residue_name_lookup(sequence)

    _exit_on_sequence_and_shift_residue_name_mismatch(
        assigned_shifts, residue_name_lookup
    )

    # TODO: filter assigned shifts which are not in the target_chain

    pseudo_atom_shifts = _build_pseudo_atom_shifts(unassigned_shifts)
    assigned_atom_shifts = _build_assigned_atom_shifts(assigned_shifts)

    lines, headers = _build_output_table(assigned_atom_shifts, pseudo_atom_shifts)

    exit_if_file_has_bytes_and_no_force(output_file, force)

    file_h = sys.stdout if output_file == STDOUT else open(output_file, "w")

    print(
        tabulate(lines, headers=headers, tablefmt="plain", floatfmt="7.3f"), file=file_h
    )

    if output_file != STDOUT:
        file_h.close()

    return None if output_file == STDOUT else entry


def _has_numbers(input: str) -> bool:
    return any(char.isdigit() for char in input)


def _build_output_table(assigned_atom_shifts, pseudo_atom_shifts):
    lines = []
    assigned_residues = _build_residues(assigned_atom_shifts)
    pseudo_residues = _build_residues(pseudo_atom_shifts)
    headers = _filter_headings_by_pseudoatoms(
        OUPUT_TABLE_HEADERS, {**pseudo_residues, **assigned_residues}
    )
    lines.extend(_build_residue_rows(assigned_atom_shifts, headers, assigned=True))
    lines.extend(_build_residue_rows(pseudo_atom_shifts, headers, assigned=False))

    return lines, headers


def _build_assigned_atom_shifts(assigned_shifts):
    assigned_atom_shifts = {}
    for shift in assigned_shifts:
        pseudo_atom = MarsAtom(
            atom_name=shift.atom.atom_name,
            sequence_code=shift.atom.residue.sequence_code,
            residue_name=shift.atom.residue.residue_name,
        )
        assigned_atom_shifts[pseudo_atom] = shift.value

    _add_m1_shifts(assigned_atom_shifts)

    return assigned_atom_shifts


def _build_residue_rows(atom_shifts, headers, assigned=False):

    residues = _build_residues(atom_shifts)
    residue_names = {
        atom_shift.sequence_code: atom_shift.residue_name for atom_shift in atom_shifts
    }
    lines = []
    for residue_num in sorted(residues):
        pseudo_atoms = residues[residue_num]

        line = []
        lines.append(line)

        if assigned:
            line.append(f"AR_{residue_num}_{residue_names[residue_num]}")
        else:
            line.append(f"PR_{residue_num}")

        for heading, heading_key in HEADINGS:

            if heading_key in pseudo_atoms:

                pseudo_atom = pseudo_atoms[heading_key]
                shift = atom_shifts[pseudo_atom]

                line.append("%-7.3f    " % shift)
            elif heading.upper() in headers:
                line.append("-         ")

    return lines


def _build_residues(assigned_atom_shifts):
    result = {}
    for assigned_atom in assigned_atom_shifts:
        key = (assigned_atom.atom_name, assigned_atom.negative_offset)
        result.setdefault(assigned_atom.sequence_code, {})[key] = assigned_atom
    return result


def _exit_on_sequence_and_shift_residue_name_mismatch(shifts, residue_name_lookup):
    for shift in shifts:
        chain_code = str(shift.atom.residue.chain_code)
        sequence_code = shift.atom.residue.sequence_code
        sequence_residue_name = get_residue_name_from_lookup(
            chain_code, sequence_code, residue_name_lookup
        )
        if sequence_residue_name != shift.atom.residue.residue_name:
            frame_line = shift.frame_line.split("\n")
            frame_headers = frame_line[0].split(",")
            frame_row = frame_line[1].split(",")
            frame_headers[0] = f"#{frame_headers[0]}"
            line_info = tabulate([frame_headers, frame_row], tablefmt="plain")
            msg = f"""
                    error for atom {shift.atom.atom_name} in chain {shift.atom.residue.chain_code}
                    the residue type in the sequence doesn't match residue type in shifts

                    sequence residue name {sequence_residue_name}
                    shifts residue name {shift.atom.residue.residue_name}

                    the shift came from
                    frame name {shift.frame_name}
                    row number {shift.frame_row}
                    data in the row

                """
            msg = dedent(msg)
            msg += line_info
            exit_error(msg)


def _build_pseudo_atom_shifts(unassigned_shifts):
    pseudo_atom_shifts = {}
    for shift in unassigned_shifts:
        # first deal with completely unassigned residues chain_code @- and sequence_code @xxxx where xxx
        # is the pseudo residue
        sequence_code = str(shift.atom.residue.sequence_code)
        atom_name = shift.atom.atom_name

        sequence_code = sequence_code.lstrip("@")
        sequence_code_fields = sequence_code.split("-")

        if len(sequence_code_fields) > 2:
            continue

        sequence_code = sequence_code_fields[0]
        if not is_int(sequence_code):
            continue
        sequence_code = int(sequence_code)

        negative_offset = (
            sequence_code_fields[1] if len(sequence_code_fields) == 2 else 0
        )

        if not is_int(negative_offset):
            continue
        else:
            negative_offset = int(negative_offset)

        if negative_offset not in (0, 1):
            continue

        atom_name = atom_name.replace("@", "")
        if _has_numbers(atom_name):
            continue

        pseudo_atom = MarsAtom(
            atom_name=atom_name,
            sequence_code=sequence_code,
            residue_name=shift.atom.residue.residue_name,
            negative_offset=negative_offset,
        )

        if not is_float(shift.value):
            continue

            # TODO: should be value
        value = float(shift.value)

        pseudo_atom_shifts[pseudo_atom] = value

    return pseudo_atom_shifts


def _filter_headings_by_pseudoatoms(base_headers, pseudo_residues):
    atom_names = set()
    for pseudo_residue in pseudo_residues.values():
        for pseudo_atom in pseudo_residue.values():
            offset = (
                f"-{pseudo_atom.negative_offset}"
                if pseudo_atom.negative_offset == 1
                else ""
            )
            atom_name = pseudo_atom.atom_name.upper()
            atom_name = "CO" if atom_name == "C" else atom_name
            atom_names.add(f"{atom_name}{offset}")
    headers = []
    for header in base_headers:
        if header == "" or header in atom_names:
            headers.append(header)
    return headers


def _add_m1_shifts(assigned_shifts):

    new_shifts = {}

    shifts_by_residue_and_atom = {}
    sequence_code_to_residue_name = {}
    for mars_atom in assigned_shifts:
        shifts_by_residue_and_atom.setdefault(mars_atom.sequence_code, {})[
            mars_atom.atom_name
        ] = mars_atom
        sequence_code_to_residue_name[mars_atom.sequence_code] = mars_atom.residue_name

    for sequence_code, atoms_and_shifts in shifts_by_residue_and_atom.items():
        residue_name = sequence_code_to_residue_name[sequence_code]
        sequence_code_m1 = sequence_code - 1
        m1_shifts = (
            shifts_by_residue_and_atom[sequence_code_m1]
            if sequence_code_m1 in shifts_by_residue_and_atom
            else {}
        )
        for atom_name in "HA CA CB C".split():
            if atom_name in m1_shifts:
                mars_atom = m1_shifts[atom_name]
                new_atom = MarsAtom(
                    atom_name=atom_name,
                    sequence_code=sequence_code,
                    residue_name=residue_name,
                    negative_offset=1,
                )
                new_shifts[new_atom] = assigned_shifts[mars_atom]

    assigned_shifts.update(new_shifts)

    return assigned_shifts



from typing import Dict, List, Tuple

import tabulate
import typer

from nef_pipelines.lib.sequence_lib import get_sequence_or_exit
from nef_pipelines.lib.structures import (
    AtomLabel,
    Linking,
    RdcRestraint,
    SequenceResidue,
)
from nef_pipelines.lib.util import chunks, exit_error, read_float_or_exit
from nef_pipelines.transcoders.pales import export_app

app = typer.Typer()


# noinspection PyUnusedLocal
@export_app.command()
def template(
    chain_code: str = typer.Option(
        "A",
        "-c",
        "--chain",
        help="chain code a single chain code",
        metavar="<CHAIN-CODE>",
    ),
    template_atoms: Tuple[str, str] = typer.Option(
        ("HN", "N"),
        "-a",
        "--atoms",
        help="for templates the atoms for the restraintsb to be between",
    ),
    # see _parse_weights for parsing and adding defaults
    raw_weights: List[str] = typer.Option(
        [],
        "-w",
        "--weight",
        help="weights for rdcs as a comma separated triple of 2 atom names and a weight [no spaces e.g HN,N,1.0], "
        "multiple weights may be added by repeated calls, efault is HN,N,1.0",
    ),
):
    """convert nef rdc restraints to pales template file for prediction of rdcs"""

    weights = _parse_weights(raw_weights)
    sequence = get_sequence_or_exit()

    restaints = _build_dummy_restraints(sequence, template_atoms)

    _print_restraints(restaints, weights=weights)


def _parse_weights(raw_weights):

    result = {}
    for raw_weight in raw_weights:

        raw_weight_fields = raw_weight.split(",")

        if len(raw_weight_fields) != 3:
            exit_error(
                f"bad weight {raw_weight} weights should have 3 fields separated by commas [no spaces] "
                f"atom_1,atom_2,weight. e.g HN,N,1.0"
            )

        atom_1, atom_2, raw_rdc_weight = raw_weight_fields

        rdc_weight = read_float_or_exit(
            raw_rdc_weight,
            message=f"weights should have 3 fields separated by commas [no spaces] atom_1,atom_2,weight weight should "
            f"be a float i got {raw_weight} for field feild [{raw_rdc_weight}] which is not a float",
        )

        key = _build_weights_key(atom_1, atom_2)
        result[key] = rdc_weight

    # is it not possible to add default weights?
    HN_N_key = _build_weights_key("HN", "N")
    if HN_N_key not in result:
        result[HN_N_key] = 1.0

    return result


def _build_weights_key(atom_1, atom_2):
    return tuple(sorted([atom_1, atom_2]))


def _build_dummy_restraints(
    sequence: SequenceResidue, atom_names: Tuple[str, str]
) -> List[RdcRestraint]:
    """
    note we use xplor names as pales is a NIH product
    :param sequence:
    :param atom_names:
    :return:
    """
    restraints = []
    for residue in sequence:

        # special case prolines, but really we should check chem comap for atoms...
        if residue.residue_name == "PRO" and "HN" in atom_names:
            continue

        if residue.linking == Linking.START and "HN" in atom_names:
            continue

        atom_1 = AtomLabel(
            SequenceResidue(
                residue.chain_code, residue.sequence_code, residue.residue_name
            ),
            atom_names[0],
        )
        atom_2 = AtomLabel(
            SequenceResidue(
                residue.chain_code, residue.sequence_code, residue.residue_name
            ),
            atom_names[1],
        )
        restraint = RdcRestraint(atom_1, atom_2, 0.0, 0.0)
        restraints.append(restraint)

    return restraints


def _print_restraints(
    restraints: List[RdcRestraint], weights: Dict[Tuple[str, str], float]
):
    VARS = "VARS   RESID_I RESNAME_I ATOMNAME_I RESID_J RESNAME_J ATOMNAME_J D DD W".split()
    FORMAT = "FORMAT %5d     %6s       %6s        %5d     %6s    %6s %9.3f %9.3f %.2f".split()

    table = [VARS, FORMAT]

    for i, restraint in enumerate(restraints):
        atom_1 = restraint.atom_1
        atom_2 = restraint.atom_2

        weights_key = _build_weights_key(atom_1.atom_name, atom_2.atom_name)
        weight = weights[weights_key]
        row = [
            "",
            atom_1.residue.sequence_code,
            atom_1.residue.residue_name,
            atom_1.atom_name,
            atom_2.residue.sequence_code,
            atom_2.residue.residue_name,
            atom_2.atom_name,
            0.000,
            0.000,
            weight,
        ]
        table.append(row)

    print(tabulate.tabulate(table, tablefmt="plain"))


def _print_pipe_sequence(sequence_1_let: List[str]):

    rows = chunks(sequence_1_let, 100)

    for row in rows:
        row_chunks = list(chunks(row, 10))
        row_strings = ["".join(chunk) for chunk in row_chunks]
        print(f'DATA SEQUENCE {" ".join(row_strings)}')
